#------------------------------------------------------------------------------------------------
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted.
#------------------------------------------------------------------------------------------------
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#------------------------------------------------------------------------------------------------
# IT WAS CREATED AND TESTED IN A WINDOWS ENVIRONMENT ONLY AND MAY NOT FUNCTION ON A MAC.
#------------------------------------------------------------------------------------------------
# THIS PLUGIN IS ANOTHER FIGMENT OF MY IMAGENATION AND IS NOT BASED ON OR COMPLY WITH ANYTHING! 
#------------------------------------------------------------------------------------------------
#		Name:	Double Line Plus
# Author:	sdmitch
#  Usage:	Create parallel lines offset at given distance of defined centerline.  Corners are rounded
#					if a positive radius is entered
#		Date:	Nov 2013
# ----------------------------------------------------------------------------------------------- #
#
require 'sketchup'
#
module SDM

	class Double_Line_Plus

		def initialize
			@ip = Sketchup::InputPoint.new
			@ip1 = Sketchup::InputPoint.new
			@aMod = Sketchup.active_model
			@aEnt = @aMod.active_entities
			Sketchup.send_action "viewTop:"
			@show_dialog = true
			@hw = '10'.to_l if !@hw
			@intrad=@hw if !@intrad
			@extrad=@intrad if !@extrad
			self.reset
		end

		def reset
			@pcnt = 0
			@cpts = []
			@lpts = []
			@rpts = []
			@ip1.clear
			@drawn = false
			if @show_dialog
				ans=UI.inputbox(["Endpunkte?:","Mitte Versatz:","Radius Int:","Radius Ext:"],["Nein",@hw,@intrad,@extrad],["Ja|Nein"], "Doppel Linie Plus Einstellung")
				if !ans
					Sketchup.send_action "selectSelectionTool:" 
					return nil
				end
				@end_caps = ans[0].to_s
				@hw=ans[1].to_l; @intrad=ans[2].to_l; @extrad=ans[3].to_l
				@hw = '5'.to_l if (@hw <= 0.0.inch); @md=@hw*2.0
				@show_dialog = false
			end
			@prompt = "Wähle Startpukt oder gib x,y oder x,y,z ein"
			Sketchup::set_status_text("WBreite", SB_VCB_LABEL)
			Sketchup::set_status_text(@hw, SB_VCB_VALUE)
			@aMod.start_operation "Dbl Line"
		end

		def onMouseMove(flags, x, y, view)
			if( !@ip.pick(view, x, y, @ip1) )
				return false
			end
			@cpts[@pcnt] = @ip.position
			@ip1.copy! @ip
			view.tooltip = @ip.tooltip     
			Sketchup::set_status_text(@prompt + "     aktuell(x,y,z)=#{@ip.position}", SB_PROMPT)
			if @pcnt > 0
				@dist = @cpts[@pcnt-1].distance @cpts[@pcnt]
				Sketchup::set_status_text "Distanz", SB_VCB_LABEL
				Sketchup::set_status_text @dist.to_s, SB_VCB_VALUE
				view.invalidate 
			end
		end

		def onLButtonDown(flags, x, y, view)
			@cpts[@pcnt] = @ip.position
			# Sketchup.active_model.entities.add_cpoint(@cpts[@pcnt])
			if @pcnt==0 || @dist > @md
				self.increment_pcnt 
			else
				UI.beep
			end
			view.lock_inference
		end
		
		def increment_pcnt
			@pcnt += 1
			case @pcnt
				when 1
					@prompt = "Wähle nächsten Punkt oder gib distanz oder x,y oder x,y,z. ein"
					Sketchup::set_status_text "", SB_VCB_LABEL
					Sketchup::set_status_text "", SB_VCB_VALUE
				when 2
					@prompt = "Wähle nächsten Punkt oder gib distanz oder x,y oder x,y,z.  Rechtsklick zum  Beenden"
					@v1 = @cpts[0].vector_to(@cpts[1]).normalize!
					az1=self.vector_to_azimuth(@v1); da=az1+90.degrees; 
					da = (da + 360.degrees)%360.degrees
					dv=Geom::Vector3d.new(Math.cos(da),Math.sin(da),0.0)
					@lpts[0]=@cpts[0].offset(dv,@hw)
					@rpts[0]=@cpts[0].offset(dv.reverse,@hw)
				else
					@v2 = @cpts[@pcnt-2].vector_to(@cpts[@pcnt-1]).normalize!
					az1=self.vector_to_azimuth(@v1); az2=self.vector_to_azimuth(@v2); dfl=self.deflection(az1,az2)
					da=az1+90.degrees+dfl/2.0; da = (da + 360.degrees)%360.degrees
					dv=Geom::Vector3d.new(Math.cos(da),Math.sin(da),0.0); hyp = @hw/Math.cos(dfl/2.0)
					@lpts[@pcnt-2]=@cpts[@pcnt-2].offset(dv,hyp)
					@rpts[@pcnt-2]=@cpts[@pcnt-2].offset(dv.reverse,hyp)
					@v1 = @v2
			end
		end
		
		def onRButtonDown(flags, x, y, view)
			@pcnt>0 ? self.dline_end : onCancel(flags,view)
		end

		def onKeyDown(key, repeat, flags, view)
			case key
				when VK_END then self.dline_end
			end
		end
		
		def onUserText(text, view)
			begin
				x,y,z=text.split(",")
				unless y
					d=x.to_l
					if d >= @md
						v=@cpts[@pcnt-1].vector_to(@cpts[@pcnt]).normalize!
						@cpts[@pcnt]=@cpts[@pcnt-1].offset(v,d)
					else
						UI.messagebox "Minimum distanz ist " + Sketchup.format_length(@md); return
					end
				else
					z=0.0 if !z
					@cpts[@pcnt]=Geom::Point3d.new(x.to_l,y.to_l,z.to_l)
				end
				self.increment_pcnt
				@cpts[@pcnt] = @ip.position
				view.invalidate
			rescue
				UI.beep
			end
		end

		def onCancel(flag, view)
			view.invalidate if @drawn
			Sketchup.send_action "selectSelectionTool:"
		end

		def draw(view)
			@drawn = false		
			if( @ip.valid? && @ip.display? )
				@ip.draw(view)
				@drawn = true
			end
			return if @pcnt == 0
			view.drawing_color = "green"
			inference_locked = view.inference_locked?
			view.line_width = 3 #if inference_locked
			view.draw(GL_LINE_STRIP, @cpts)
			view.line_width = 1 #if inference_locked
			@drawn = true
			return if @pcnt < 3
			view.drawing_color = "magenta"
			view.line_width = 3 #if inference_locked
			view.draw(GL_LINE_STRIP, @rpts)
			view.draw(GL_LINE_STRIP, @lpts)
			view.line_width = 1 #if inference_locked
		end

		def dline_end
			@cpts[@pcnt] = @ip.position
			# Sketchup.active_model.entities.add_cpoint(@cpts[@pcnt])
			@v2 = (@cpts[@pcnt]-@cpts[@pcnt-1]).normalize!
			az1=self.vector_to_azimuth(@v1); az2=self.vector_to_azimuth(@v2); dfl=self.deflection(az1,az2)
			da=az1+90.degrees+dfl/2.0; da = (da + 360.degrees)%360.degrees
			dv=Geom::Vector3d.new(Math.cos(da),Math.sin(da),0.0); hyp = @hw/Math.cos(dfl/2.0)
			@lpts[@pcnt-1]=@cpts[@pcnt-1].offset(dv,hyp)
			@rpts[@pcnt-1]=@cpts[@pcnt-1].offset(dv.reverse,hyp)
			az2=self.vector_to_azimuth(@v2); da=az2+90.degrees; da = (da + 360.degrees)%360.degrees
			dv=Geom::Vector3d.new(Math.cos(da),Math.sin(da),0.0)
			@lpts[@pcnt]=@cpts[@pcnt].offset(dv,@hw)
			@rpts[@pcnt]=@cpts[@pcnt].offset(dv.reverse,@hw)
			@edges=@aEnt.add_edges(@lpts); @left=true; self.add_fillets if (@intrad > 0 or @extrad > 0)
			@edges=@aEnt.add_edges(@rpts); @left=false; self.add_fillets if (@intrad > 0 or @extrad > 0)
			if @end_caps == "Yes" then
				@aEnt.add_line(@lpts[0],@rpts[0])
				@aEnt.add_line(@lpts[-1],@rpts[-1])
			end			
			@aMod.commit_operation
			@show_dialog = true
			self.reset
		end
		
		def  vector_to_azimuth(v)
		# --------------------------------------------------------------------------------
		#	Compute Azimuth based on positive x-axis as 0 degrees
		# --------------------------------------------------------------------------------
			v.normalize!
			va=Math::acos(v.z) # Vertical angle
			if v.x < 0
				az = Math::PI - Math::asin(v.y) # Quadrant III or IV
			elsif v.y < 0
				az = Math::PI * 2 + Math::asin(v.y) # Quadrant II
			else
				az = Math::asin(v.y) # Quadrant I
			end
			return az # Horizontal angle
		end
		
		def deflection(az1,az2)
		#  --------------------------------------------------------------------  #
		#	Compute the Deflection angle between given azimuths
		#  --------------------------------------------------------------------  #
			defl=az2-az1
			defl += Math::PI * 2 if defl < -Math::PI
			defl -= Math::PI * 2 if defl >   Math::PI
			return defl
		end
		
		def add_fillets
				if @edges.length>=2
					pts=[]
					for i in 1...@edges.length
						poi=Geom.intersect_line_line @edges[i-1].line,@edges[i].line
						if poi
							p0,v0=@edges[i-1].line; p1,v1=@edges[i].line
							az0=self.vector_to_azimuth(v0); az1=self.vector_to_azimuth(v1); dfl=self.deflection(az0,az1)
							if @left
								dfl>=0.0 ? @rad=@intrad : @rad=@extrad
							else
								dfl<0.0 ? @rad=@intrad : @rad=@extrad
							end
							if @rad > 0.0
								v0.reverse! unless p0==poi
								v1.reverse! unless p1==poi
								ang=(v0.angle_between(v1))/2;
								hyp=@rad/Math.sin(ang); td=hyp*Math.cos(ang)
								v2c=Geom::Vector3d.linear_combination(0.5,v0,0.5,v1)
								ctr=poi.offset(v2c,hyp)
								pot0=poi.offset(v0,td);pot1=poi.offset(v1,td)
								sv=ctr.vector_to(pot0).normalize
								ev=ctr.vector_to(pot1).normalize
								ang=sv.angle_between(ev); seg=[(ang/15.degrees).ceil,6].max
								@aEnt.add_arc ctr,sv,sv*ev,@rad,0,ang,seg
								pts<<[poi,pot0,pot1]
							end
						end
					end
					for i in 0...pts.length
						line=@aEnt.add_line(pts[i][0],pts[i][1]); line.erase!
						line=@aEnt.add_line(pts[i][0],pts[i][2]); line.erase!
					end
				end
		end
		
	end # of class DoubleLine

end

# ------------------  MENU SETUP ---------------------- #
unless $sdm_tools_menu
	$sdm_tools_menu = UI.menu("Plugins").add_submenu("Werkzeuge von sdmitch")
	$sdm_Edge_tools = $sdm_tools_menu.add_submenu("Kanten und Linien")
	$sdm_Face_tools = $sdm_tools_menu.add_submenu("Flächen Werkzeuge")
	$sdm_CorG_tools = $sdm_tools_menu.add_submenu("Gruppen und Komponenten")
	$sdm_Misc_tools = $sdm_tools_menu.add_submenu("Diverse Werkzeuge")
end
unless file_loaded?(__FILE__)
	$sdm_Edge_tools.add_item("Doppel Linie Plus") { Sketchup.active_model.select_tool SDM::Double_Line_Plus.new }
	file_loaded(__FILE__)
end
# ------------------------------------------------------ #
